#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import os
import pathlib

from unittest import TestCase

from mbed_tools.project import MbedProgram
from mbed_tools.project.exceptions import ExistingProgram, ProgramNotFound, MbedOSNotFound
from mbed_tools.project.mbed_program import _find_program_root, parse_url
from mbed_tools.project._internal.project_data import MbedProgramFiles
from tests.project.factories import make_mbed_program_files, make_mbed_os_files, patchfs


DEFAULT_BUILD_SUBDIR = pathlib.Path("K64F", "develop", "GCC_ARM")


def from_new_set_target_toolchain(program_root):
    program = MbedProgram.from_new(program_root)
    parent_build_dir = program.files.cmake_build_dir
    program.files.cmake_build_dir = parent_build_dir / DEFAULT_BUILD_SUBDIR
    return program


class TestInitialiseProgram(TestCase):
    @patchfs
    def test_from_new_local_dir_raises_if_path_is_existing_program(self, fs):
        program_root = pathlib.Path(fs, "programfoo")
        program_root.mkdir()
        (program_root / "mbed-os.lib").touch()

        with self.assertRaises(ExistingProgram):
            MbedProgram.from_new(program_root)

    @patchfs
    def test_from_new_local_dir_generates_valid_program_creating_directory(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        fs_root.mkdir()
        program_root = fs_root / "programfoo"

        program = from_new_set_target_toolchain(program_root)

        self.assertEqual(program.files, MbedProgramFiles.from_existing(program_root, DEFAULT_BUILD_SUBDIR))

    @patchfs
    def test_from_new_local_dir_generates_valid_program_creating_directory_in_cwd(self, fs):
        old_cwd = os.getcwd()
        try:
            fs_root = pathlib.Path(fs, "foo")
            fs_root.mkdir()
            os.chdir(fs_root)
            program_root = pathlib.Path("programfoo")

            program = from_new_set_target_toolchain(program_root)

            self.assertEqual(program.files, MbedProgramFiles.from_existing(program_root, DEFAULT_BUILD_SUBDIR))
        finally:
            os.chdir(old_cwd)

    @patchfs
    def test_from_new_local_dir_generates_valid_program_existing_directory(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        fs_root.mkdir()
        program_root = fs_root / "programfoo"
        program_root.mkdir()

        program = from_new_set_target_toolchain(program_root)

        self.assertEqual(program.files, MbedProgramFiles.from_existing(program_root, DEFAULT_BUILD_SUBDIR))

    @patchfs
    def test_from_existing_raises_if_path_is_not_a_program(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        fs_root.mkdir()
        program_root = fs_root / "programfoo"

        with self.assertRaises(ProgramNotFound):
            MbedProgram.from_existing(program_root, DEFAULT_BUILD_SUBDIR)

    @patchfs
    def test_from_existing_raises_if_no_mbed_os_dir_found_and_check_mbed_os_is_true(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        make_mbed_program_files(fs_root)

        with self.assertRaises(MbedOSNotFound):
            MbedProgram.from_existing(fs_root, DEFAULT_BUILD_SUBDIR, check_mbed_os=True)

    @patchfs
    def test_from_existing_returns_valid_program(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        make_mbed_program_files(fs_root)
        make_mbed_os_files(fs_root / "mbed-os")

        program = MbedProgram.from_existing(fs_root, DEFAULT_BUILD_SUBDIR)

        self.assertTrue(program.files.app_config_file.exists())
        self.assertTrue(program.mbed_os.root.exists())

    @patchfs
    def test_from_existing_with_mbed_os_path_returns_valid_program(self, fs):
        fs_root = pathlib.Path(fs, "foo")
        mbed_os_path = fs_root / "extern/mbed-os"
        mbed_os_path.mkdir(parents=True)
        make_mbed_program_files(fs_root)
        make_mbed_os_files(mbed_os_path)

        program = MbedProgram.from_existing(fs_root, DEFAULT_BUILD_SUBDIR, mbed_os_path)

        self.assertTrue(program.files.app_config_file.exists())
        self.assertTrue(program.mbed_os.root.exists())


class TestParseURL(TestCase):
    def test_creates_url_and_dst_dir_from_name(self):
        name = "mbed-os-example-blows-up-board"
        data = parse_url(name)

        self.assertEqual(data["url"], f"https://github.com/armmbed/{name}")
        self.assertEqual(data["dst_path"], name)

    def test_creates_valid_dst_dir_from_url(self):
        url = "https://superversioncontrol/superorg/mbed-os-example-numskull"
        data = parse_url(url)

        self.assertEqual(data["url"], url)
        self.assertEqual(data["dst_path"], "mbed-os-example-numskull")

    def test_creates_valid_dst_dir_from_ssh_url(self):
        url = "git@superversioncontrol:superorg/mbed-os-example-numskull"
        data = parse_url(url)
        self.assertEqual(data["url"], url)
        self.assertEqual(data["dst_path"], "mbed-os-example-numskull")


class TestFindProgramRoot(TestCase):
    @patchfs
    def test_finds_program_higher_in_dir_tree(self, fs):
        program_root = pathlib.Path(fs, "foo")
        pwd = program_root / "subprojfoo" / "libbar"
        make_mbed_program_files(program_root)
        pwd.mkdir(parents=True)

        self.assertEqual(_find_program_root(pwd), program_root.resolve())

    @patchfs
    def test_finds_program_at_current_path(self, fs):
        program_root = pathlib.Path(fs, "foo")
        make_mbed_program_files(program_root)

        self.assertEqual(_find_program_root(program_root), program_root.resolve())

    @patchfs
    def test_raises_if_no_program_found(self, fs):
        program_root = pathlib.Path(fs, "foo")
        program_root.mkdir()

        with self.assertRaises(ProgramNotFound):
            _find_program_root(program_root)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Helpers for python language related functions."""
from typing import Iterable, List


def flatten_nested(input_iter: Iterable) -> List:
    """Flatten a nested Iterable with arbitrary levels of nesting.

    If the input is an iterator then this function will exhaust it.

    Args:
        input_iter: The input Iterable which may or may not be nested.

    Returns:
        A flat list created from the input_iter.
        If input_iter has no nesting its elements are appended to a list and returned.
    """
    output = []
    for elem in input_iter:
        if isinstance(elem, Iterable) and not isinstance(elem, str):
            output += flatten_nested(elem)
        else:
            output.append(elem)

    return output



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Render jinja templates required by the project package."""
import datetime

from pathlib import Path

import jinja2

TEMPLATES_DIRECTORY = Path("_internal", "templates")


def render_cmakelists_template(cmakelists_file: Path, program_name: str) -> None:
    """Render CMakeLists.tmpl with the copyright year and program name as the app target name.

    Args:
        cmakelists_file: The path where CMakeLists.txt will be written.
        program_name: The name of the program, will be used as the app target name.
    """
    cmakelists_file.write_text(
        render_jinja_template(
            "CMakeLists.tmpl", {"program_name": program_name, "year": str(datetime.datetime.now().year)}
        )
    )


def render_main_cpp_template(main_cpp: Path) -> None:
    """Render a basic main.cpp which prints a hello message and returns.

    Args:
        main_cpp: Path where the main.cpp file will be written.
    """
    main_cpp.write_text(render_jinja_template("main.tmpl", {"year": str(datetime.datetime.now().year)}))


def render_gitignore_template(gitignore: Path) -> None:
    """Write out a basic gitignore file ignoring the build and config directory.

    Args:
        gitignore: The path where the gitignore file will be written.
    """
    gitignore.write_text(render_jinja_template("gitignore.tmpl", {}))


def render_jinja_template(template_name: str, context: dict) -> str:
    """Render a jinja template.

    Args:
        template_name: The name of the template being rendered.
        context: Data to render into the jinja template.
    """
    env = jinja2.Environment(loader=jinja2.PackageLoader("mbed_tools.project", str(TEMPLATES_DIRECTORY)))
    template = env.get_template(template_name)
    return template.render(context)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json

import pytest

from mbed_tools.lib.json_helpers import decode_json_file


def test_invalid_json(tmp_path):
    lib_json_path = tmp_path / "mbed_lib.json"
    lib_json_path.write_text("name")

    with pytest.raises(json.JSONDecodeError):
        decode_json_file(lib_json_path)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

import pytest

from unittest import mock

from mbed_tools.project import initialise_project, import_project, deploy_project, get_known_libs


@pytest.fixture
def mock_libs():
    with mock.patch("mbed_tools.project.project.LibraryReferences") as libs:
        yield libs


@pytest.fixture
def mock_program():
    with mock.patch("mbed_tools.project.project.MbedProgram") as prog:
        yield prog


@pytest.fixture
def mock_git():
    with mock.patch("mbed_tools.project.project.git_utils") as gutils:
        yield gutils


class TestInitialiseProject:
    def test_fetches_mbed_os_when_create_only_is_false(self, mock_libs, mock_program):
        path = pathlib.Path()
        initialise_project(path, create_only=False)

        mock_program.from_new.assert_called_once_with(path)
        mock_libs().fetch.assert_called_once()

    def test_skips_mbed_os_when_create_only_is_true(self, mock_libs, mock_program):
        path = pathlib.Path()
        initialise_project(path, create_only=True)

        mock_program.from_new.assert_called_once_with(path)
        mock_libs().fetch.assert_not_called()


class TestImportProject:
    def test_clones_from_remote(self, mock_git):
        url = "https://git.com/gitorg/repo"
        import_project(url, recursive=False)

        mock_git.clone.assert_called_once_with(url, pathlib.Path(url.rsplit("/", maxsplit=1)[-1]))

    def test_resolves_libs_when_recursive_is_true(self, mock_git, mock_libs):
        url = "https://git.com/gitorg/repo"
        import_project(url, recursive=True)

        mock_git.clone.assert_called_once_with(url, pathlib.Path(url.rsplit("/", maxsplit=1)[-1]))
        mock_libs().fetch.assert_called_once()


class TestDeployProject:
    def test_checks_out_libraries(self, mock_libs):
        path = pathlib.Path("somewhere")
        deploy_project(path, force=False)

        mock_libs().checkout.assert_called_once_with(force=False)

    def test_resolves_libs_if_unresolved_detected(self, mock_libs):
        mock_libs().iter_unresolved.return_value = [1]
        path = pathlib.Path("somewhere")
        deploy_project(path)

        mock_libs().fetch.assert_called_once()


class TestPrintLibs:
    def test_list_libraries_gets_known_lib_list(self, mock_libs):
        path = pathlib.Path("somewhere")
        mock_libs().iter_resolved.return_value = ["", ""]

        libs = get_known_libs(path)

        assert libs == ["", ""]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from tests.devices.markers import windows_only


@windows_only
class TestDiskUid(unittest.TestCase):
    def test_disk_uid_parsing(self):
        from mbed_tools.devices._internal.windows.disk_drive import Win32DiskIdParser, WindowsUID

        # Tests the parsing of real data.
        # ST Link
        pnpid1 = "USBSTOR\\DISK&VEN_MBED&PROD_MICROCONTROLLER&REV_1.0\\9&175DDF0B&0&066DFF555654725187095153&0"
        serial_number1 = "066DFF555654725187095153"
        self.assertEqual(
            Win32DiskIdParser().parse(pnpid=pnpid1, serial_number=serial_number1),
            WindowsUID(
                uid="066dff555654725187095153",
                raw_uid="9&175DDF0B&0&066DFF555654725187095153&0",
                serial_number="066DFF555654725187095153",
            ),
        )
        pnpid2 = "USBSTOR\\DISK&VEN_MBED&PROD_MICROCONTROLLER&REV_1.0\\9&3849C7A8&0&0672FF574953867567051035&0"
        serial_number2 = "0672FF574953867567051035"
        self.assertEqual(
            Win32DiskIdParser().parse(pnpid=pnpid2, serial_number=serial_number2),
            WindowsUID(
                uid="0672ff574953867567051035",
                raw_uid="9&3849C7A8&0&0672FF574953867567051035&0",
                serial_number="0672FF574953867567051035",
            ),
        )
        # System disk
        pnpid3 = "SCSI\\DISK&VEN_SAMSUNG&PROD_MZNLN512HMJP-000\\4&143821B1&0&000100"
        serial_number3 = "S2XANX0J211020"
        self.assertEqual(
            Win32DiskIdParser().parse(pnpid=pnpid3, serial_number=serial_number3),
            WindowsUID(uid="4&143821b1&0&000100", raw_uid="4&143821B1&0&000100", serial_number="S2XANX0J211020"),
        )
        # Daplink
        pnpid4 = "USBSTOR\\DISK&VEN_MBED&PROD_VFS&REV_0.1\\0240000034544E45001A00018AA900292011000097969900&0"
        serial_number4 = "0240000034544E45001A00018AA900292011000097969900"
        self.assertEqual(
            Win32DiskIdParser().parse(pnpid=pnpid4, serial_number=serial_number4),
            WindowsUID(
                uid="0240000034544e45001a00018aa900292011000097969900",
                raw_uid="0240000034544E45001A00018AA900292011000097969900&0",
                serial_number="0240000034544E45001A00018AA900292011000097969900",
            ),
        )
        # J Link
        pnpid5 = "USBSTOR\\DISK&VEN_SEGGER&PROD_MSD_VOLUME&REV_1.00\\9&DBDECF6&0&000440112138&0"
        serial_number5 = "                        134657890"
        self.assertEqual(
            Win32DiskIdParser().parse(pnpid=pnpid5, serial_number=serial_number5),
            WindowsUID(
                uid="000440112138",
                raw_uid="9&DBDECF6&0&000440112138&0",
                serial_number="                         134657890",
            ),
        )

    def test_uid_linking_between_usb_and_disk(self):
        from mbed_tools.devices._internal.windows.usb_device_identifier import UsbIdentifier, WindowsUID

        disk_uid = WindowsUID(
            uid="000440112138",
            raw_uid="9&DBDECF6&0&000440112138&0",
            serial_number="                         134657890",
        )

        usb_uid = UsbIdentifier(
            UID=WindowsUID(uid="000440112138", raw_uid="000440112138", serial_number="8&2f125ec6&0"),
            VID="1366",
            PID="1015",
            REV=None,
            MI=None,
        )
        self.assertEqual(disk_uid.presumed_serial_number, usb_uid.uid.presumed_serial_number)
        self.assertEqual(usb_uid.uid.presumed_serial_number, disk_uid.presumed_serial_number)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to launch a serial terminal to a connected Mbed device."""
from typing import Any, Optional

import click

from mbed_tools.cli.build import _get_target_id
from mbed_tools.devices import find_connected_device, get_connected_devices
from mbed_tools.devices.exceptions import MbedDevicesError
from mbed_tools.sterm import terminal


@click.command(
    help="Open a serial terminal to a connected Mbed Enabled device, or connect to a user-specified COM port."
)
@click.option(
    "-p",
    "--port",
    type=str,
    help="Communication port. Default: auto-detect. Specifying this will also ignore the -m/--mbed-target option.",
)
@click.option("-b", "--baudrate", type=int, default=9600, show_default=True, help="Communication baudrate.")
@click.option(
    "-e",
    "--echo",
    default="on",
    show_default=True,
    type=click.Choice(["on", "off"], case_sensitive=False),
    help="Switch local echo on/off.",
)
@click.option("-m", "--mbed-target", type=str, help="Mbed target to detect. Example: K64F, NUCLEO_F401RE, NRF51822...")
def sterm(port: str, baudrate: int, echo: str, mbed_target: str) -> None:
    """Launches a serial terminal to a connected device."""
    if port is None:
        port = _find_target_serial_port_or_default(mbed_target)

    terminal.run(port, baudrate, echo=True if echo == "on" else False)


def _get_connected_mbed_devices() -> Any:
    connected_devices = get_connected_devices()
    if not connected_devices.identified_devices:
        raise MbedDevicesError("No Mbed enabled devices found.")

    return connected_devices.identified_devices


def _find_target_serial_port_or_default(target: Optional[str]) -> Any:
    if target is None:
        # just return the first valid device found
        device, *_ = _get_connected_mbed_devices()
    else:
        target_name, target_id = _get_target_id(target)
        device = find_connected_device(target_name.upper(), target_id)
    return device.serial_port



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""API for listing devices."""

from operator import attrgetter
from typing import List, Optional

from mbed_tools.targets.exceptions import MbedTargetsError
from mbed_tools.targets import Board

from mbed_tools.devices._internal.detect_candidate_devices import detect_candidate_devices
from mbed_tools.devices._internal.resolve_board import resolve_board
from mbed_tools.devices._internal.exceptions import NoBoardForCandidate

from mbed_tools.devices.device import ConnectedDevices, Device
from mbed_tools.devices.exceptions import DeviceLookupFailed, NoDevicesFound


def get_connected_devices() -> ConnectedDevices:
    """Returns Mbed Devices connected to host computer.

    Connected devices which have been identified as Mbed Boards and also connected devices which are potentially
    Mbed Boards (but not could not be identified in the database) are returned.

    Raises:
        DeviceLookupFailed: If there is a problem with the process of identifying Mbed Boards from connected devices.
    """
    connected_devices = ConnectedDevices()

    board: Optional["Board"]
    for candidate_device in detect_candidate_devices():
        try:
            board = resolve_board(candidate_device)
        except NoBoardForCandidate:
            board = None
        except MbedTargetsError as err:
            raise DeviceLookupFailed("A problem occurred when looking up board data for connected devices.") from err

        connected_devices.add_device(candidate_device, board)

    return connected_devices


def find_connected_device(target_name: str, identifier: Optional[int] = None) -> Device:
    """Find a connected device matching the given target_name, if there is only one.

    Args:
        target_name: The Mbed target name of the device.
        identifier: Where multiple of the same Mbed device are connected, the associated [id].

    Raise:
        DeviceLookupFailed: Could not find device matching target_name.

    Returns:
        The first Device found matching target_name.
    """
    devices = find_all_connected_devices(target_name)
    if identifier is None and len(devices) == 1:
        return devices[0]
    elif identifier is not None and len(devices) > identifier:
        return devices[identifier]

    detected_targets = "\n".join(
        f"target: {dev.mbed_board.board_type}[{i}]," f" port: {dev.serial_port}, mount point(s): {dev.mount_points}"
        for i, dev in enumerate(devices)
    )
    if identifier is None:
        msg = (
            f"`Multiple matching, please select a connected target with [n] identifier.\n"
            f"The following {target_name}s were detected:\n{detected_targets}"
        )
    else:
        msg = (
            f"`{target_name}[{identifier}]` is not a valid connected target.\n"
            f"The following {target_name}s were detected:\n{detected_targets}"
        )
    raise DeviceLookupFailed(msg)


def find_all_connected_devices(target_name: str) -> List[Device]:
    """Find all connected devices matching the given target_name.

    Args:
        target_name: The Mbed target name of the device.

    Raises:
        NoDevicesFound: Could not find any connected devices.
        DeviceLookupFailed: Could not find a connected device matching target_name.

    Returns:
        List of Devices matching target_name.
    """
    connected = get_connected_devices()
    if not connected.identified_devices:
        raise NoDevicesFound("No Mbed enabled devices found.")

    matching_devices = sorted(
        [device for device in connected.identified_devices if device.mbed_board.board_type == target_name.upper()],
        key=attrgetter("serial_number"),
    )
    if matching_devices:
        return matching_devices

    detected_targets = "\n".join(
        f"target: {dev.mbed_board.board_type}, port: {dev.serial_port}, mount point(s): {dev.mount_points}"
        for dev in connected.identified_devices
    )
    msg = (
        f"Target '{target_name}' was not detected.\n"
        "Check the device is connected by USB, and that the name is entered correctly.\n"
        f"The following devices were detected:\n{detected_targets}"
    )
    raise DeviceLookupFailed(msg)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets._internal.board_database`."""

from unittest import mock

import logging
import pytest

import mbed_tools.targets._internal.board_database as board_database
from mbed_tools.targets.env import env


class TestGetOnlineBoardData:
    """Tests for the method `board_database.get_online_board_data`."""

    def test_401(self, caplog, requests_mock):
        """Given a 401 error code, BoardAPIError is raised."""
        caplog.set_level(logging.DEBUG)
        requests_mock.get(board_database._BOARD_API, status_code=401, text="Who are you?")
        with pytest.raises(board_database.BoardAPIError):
            board_database.get_online_board_data()
        assert any(
            x for x in caplog.records if x.levelno == logging.WARNING and "MBED_API_AUTH_TOKEN" in x.msg
        ), "Auth token should be mentioned"
        assert any(
            x for x in caplog.records if x.levelno == logging.DEBUG and "Who are you?" in x.msg
        ), "Message content should be in the debug message"

    def test_404(self, caplog, requests_mock):
        """Given a 404 error code, TargetAPIError is raised."""
        caplog.set_level(logging.DEBUG)
        requests_mock.get(board_database._BOARD_API, status_code=404, text="Not Found")
        with pytest.raises(board_database.BoardAPIError):
            board_database.get_online_board_data()
        assert any(
            x for x in caplog.records if x.levelno == logging.WARNING and "404" in x.msg
        ), "HTTP status code should be mentioned"
        assert any(
            x for x in caplog.records if x.levelno == logging.DEBUG and "Not Found" in x.msg
        ), "Message content should be in the debug message"

    def test_200_invalid_json(self, caplog, requests_mock):
        """Given a valid response but invalid json, JSONDecodeError is raised."""
        caplog.set_level(logging.DEBUG)
        requests_mock.get(board_database._BOARD_API, text="some text")
        with pytest.raises(board_database.ResponseJSONError):
            board_database.get_online_board_data()
        assert any(
            x for x in caplog.records if x.levelno == logging.WARNING and "Invalid JSON" in x.msg
        ), "Invalid JSON should be mentioned"
        assert any(
            x for x in caplog.records if x.levelno == logging.DEBUG and "some text" in x.msg
        ), "Message content should be in the debug message"

    def test_200_no_data_field(self, caplog, requests_mock):
        """Given a valid response but no data field, ResponseJSONError is raised."""
        caplog.set_level(logging.DEBUG)
        requests_mock.get(board_database._BOARD_API, json={"notdata": [], "stillnotdata": {}})
        with pytest.raises(board_database.ResponseJSONError):
            board_database.get_online_board_data()
        assert any(
            x for x in caplog.records if x.levelno == logging.WARNING and "missing the 'data' field" in x.msg
        ), "Data field should be mentioned"
        assert any(
            x for x in caplog.records if x.levelno == logging.DEBUG and "notdata, stillnotdata" in x.msg
        ), "JSON keys from message should be in the debug message"

    def test_200_value_data(self, requests_mock):
        """Given a valid response, target data is set from the returned json."""
        requests_mock.get(board_database._BOARD_API, json={"data": 42})
        board_data = board_database.get_online_board_data()
        assert 42 == board_data, "Target data should match the contents of the target API data"

    @mock.patch("mbed_tools.targets._internal.board_database.requests")
    @mock.patch("mbed_tools.targets._internal.board_database.env", spec_set=env)
    def test_auth_header_set_with_token(self, env, requests):
        """Given an authorization token env variable, get is called with authorization header."""
        env.MBED_API_AUTH_TOKEN = "token"
        header = {"Authorization": "Bearer token"}
        board_database._get_request()
        requests.get.assert_called_once_with(board_database._BOARD_API, headers=header)

    @mock.patch("mbed_tools.targets._internal.board_database.requests")
    def test_no_auth_header_set_with_empty_token_var(self, requests):
        """Given no authorization token env variable, get is called with no header."""
        board_database._get_request()
        requests.get.assert_called_once_with(board_database._BOARD_API, headers=None)

    @mock.patch("mbed_tools.targets._internal.board_database.requests.get")
    def test_logs_no_warning_on_success(self, get, caplog):
        board_database._get_request()
        assert not caplog.records

    @mock.patch("mbed_tools.targets._internal.board_database.requests.get")
    def test_raises_tools_error_on_connection_error(self, get, caplog):
        get.side_effect = board_database.requests.exceptions.ConnectionError
        with pytest.raises(board_database.BoardAPIError):
            board_database._get_request()
        assert "Unable to connect" in caplog.text
        assert len(caplog.records) == 1

    @mock.patch("mbed_tools.targets._internal.board_database.requests.get")
    def test_logs_error_on_requests_ssl_error(self, get, caplog):
        get.side_effect = board_database.requests.exceptions.SSLError
        with pytest.raises(board_database.BoardAPIError):
            board_database._get_request()
        assert "verify an SSL" in caplog.text

    @mock.patch("mbed_tools.targets._internal.board_database.requests.get")
    def test_logs_error_on_requests_proxy_error(self, get, caplog):
        get.side_effect = board_database.requests.exceptions.ProxyError
        with pytest.raises(board_database.BoardAPIError):
            board_database._get_request()
        assert "connect to proxy" in caplog.text

    @mock.patch.dict("os.environ", {"http_proxy": "http://proxy:8080", "https_proxy": "https://proxy:8080"})
    def test_requests_uses_proxy_variables(self, requests_mock):
        requests_mock.get(board_database._BOARD_API)
        board_database._get_request()
        assert requests_mock.last_request.proxies == {"http": "http://proxy:8080", "https": "https://proxy:8080"}

    @mock.patch.dict("os.environ", {"http_proxy": "http://proxy:8080", "https_proxy": "https://proxy:8080"})
    def test_raises_proxy_error_with_invalid_proxy(self, caplog):
        with pytest.raises(board_database.BoardAPIError):
            board_database._get_request()
        assert "connect to proxy" in caplog.text


class TestGetOfflineTargetData:
    """Tests for the method get_offline_target_data."""

    def test_local_target_database_file_found(self):
        """Test local database is found and loaded."""
        data = board_database.get_offline_board_data()
        assert len(data), "Some data should be returned from local database file."

    @mock.patch("mbed_tools.targets._internal.board_database.get_board_database_path")
    def test_raises_on_invalid_json(self, mocked_get_file):
        """Test raises an error when the file contains invalid JSON."""
        invalid_json = "None"
        path_mock = mock.Mock()
        path_mock.read_text.return_value = invalid_json
        mocked_get_file.return_value = path_mock
        with pytest.raises(board_database.ResponseJSONError):
            board_database.get_offline_board_data()


class TestGetLocalTargetDatabaseFile:
    def test_returns_path_to_targets(self):
        path = board_database.get_board_database_path()
        assert path.exists(), "Path to boards should exist in the package data folder."



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Objects for library reference handling."""
import logging

from dataclasses import dataclass
from pathlib import Path
from typing import Generator, List

from mbed_tools.project._internal import git_utils
from mbed_tools.project.exceptions import VersionControlError

logger = logging.getLogger(__name__)


@dataclass(frozen=True, order=True)
class MbedLibReference:
    """Metadata associated with an Mbed library.

    An Mbed library is an external dependency of an MbedProgram. The MbedProgram is made aware of the library
    dependency by the presence of a .lib file in the project tree, which we refer to as a library reference file. The
    library reference file contains a URI where the dependency's source code can be fetched.

    Attributes:
        reference_file: Path to the .lib reference file for this library.
        source_code_path: Path to the source code if it exists in the local project.
    """

    reference_file: Path
    source_code_path: Path

    def is_resolved(self) -> bool:
        """Determines if the source code for this library is present in the source tree."""
        return self.source_code_path.exists() and self.source_code_path.is_dir()

    def get_git_reference(self) -> git_utils.GitReference:
        """Get the source code location from the library reference file.

        Returns:
            Data structure containing the contents of the library reference file.
        """
        raw_ref = self.reference_file.read_text().strip()
        url, sep, ref = raw_ref.partition("#")

        if url.endswith("/"):
            url = url[:-1]

        return git_utils.GitReference(repo_url=url, ref=ref)


@dataclass
class LibraryReferences:
    """Manages library references in an MbedProgram."""

    root: Path
    ignore_paths: List[str]

    def fetch(self) -> None:
        """Recursively clone all dependencies defined in .lib files."""
        for lib in self.iter_unresolved():
            git_ref = lib.get_git_reference()
            logger.info(f"Resolving library reference {git_ref.repo_url}.")
            _clone_at_ref(git_ref.repo_url, lib.source_code_path, git_ref.ref)

        # Check if we find any new references after cloning dependencies.
        if list(self.iter_unresolved()):
            self.fetch()

    def checkout(self, force: bool) -> None:
        """Check out all resolved libs to revision specified in .lib files."""
        for lib in self.iter_resolved():
            repo = git_utils.get_repo(lib.source_code_path)
            git_ref = lib.get_git_reference()

            if not git_ref.ref:
                git_ref.ref = git_utils.get_default_branch(repo)

            git_utils.fetch(repo, git_ref.ref)
            git_utils.checkout(repo, "FETCH_HEAD", force=force)

    def iter_all(self) -> Generator[MbedLibReference, None, None]:
        """Iterate all library references in the tree.

        Yields:
            Iterator to library reference.
        """
        for lib in self.root.rglob("*.lib"):
            if not self._in_ignore_path(lib):
                yield MbedLibReference(lib, lib.with_suffix(""))

    def iter_unresolved(self) -> Generator[MbedLibReference, None, None]:
        """Iterate all unresolved library references in the tree.

        Yields:
            Iterator to library reference.
        """
        for lib in self.iter_all():
            if not lib.is_resolved():
                yield lib

    def iter_resolved(self) -> Generator[MbedLibReference, None, None]:
        """Iterate all resolved library references in the tree.

        Yields:
            Iterator to library reference.
        """
        for lib in self.iter_all():
            if lib.is_resolved():
                yield lib

    def _in_ignore_path(self, lib_reference_path: Path) -> bool:
        """Check if a library reference is in a path we want to ignore."""
        return any(p in lib_reference_path.parts for p in self.ignore_paths)


def _clone_at_ref(url: str, path: Path, ref: str) -> None:
    if ref:
        logger.info(f"Checking out revision {ref} for library {url}.")
        try:
            git_utils.clone(url, path, ref)
        except VersionControlError:
            # We weren't able to clone. Try again without the ref.
            repo = git_utils.clone(url, path)
            # We couldn't clone the ref and had to fall back to cloning
            # just the default branch. Fetch the ref before checkout, so
            # that we have it available locally.
            logger.warning(f"No tag or branch with name {ref}. Fetching full repository.")
            git_utils.fetch(repo, ref)
            git_utils.checkout(repo, "FETCH_HEAD")
    else:
        git_utils.clone(url, path)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

from click.testing import CliRunner

from mbed_tools.cli.configure import configure


class TestConfigureCommand(TestCase):
    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_generate_config_called_with_correct_arguments(self, program, generate_config):
        CliRunner().invoke(configure, ["-m", "k64f", "-t", "gcc_arm"])

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program.from_existing())

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_generate_config_called_with_mbed_os_path(self, program, generate_config):
        CliRunner().invoke(configure, ["-m", "k64f", "-t", "gcc_arm", "--mbed-os-path", "./extern/mbed-os"])

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program.from_existing())



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Representation of an Mbed-Enabled Development Board and related utilities."""
from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True, order=True)
class Board:
    """Representation of an Mbed-Enabled Development Board.

    Attributes:
        board_type: Type of board in format that allows cross-referencing with target definitions.
        board_name: Human readable name.
        product_code: Uniquely identifies a board for the online compiler.
        target_type: A confusing term that is not related to 'target' in other parts of the code.
        Identifies if a board in the online database is a `module` or a `platform` (modules are more
        complex and often have supplementary sensors, displays etc. A platform is a normal development board).
        slug: Used with target_type to identify a board's page on the mbed website.
        build_variant: Available build variants for the board.
        Can be used in conjunction with board_type for referencing targets.
        mbed_os_support: The versions of Mbed OS supported.
        mbed_enabled: Whether Mbed OS is supported or not.
    """

    board_type: str
    board_name: str
    product_code: str
    target_type: str
    slug: str
    build_variant: Tuple[str, ...]
    mbed_os_support: Tuple[str, ...]
    mbed_enabled: Tuple[str, ...]

    @classmethod
    def from_online_board_entry(cls, board_entry: dict) -> "Board":
        """Create a new instance of Board from an online database entry.

        Args:
            board_entry: A single entity retrieved from the board database API.
        """
        board_attrs = board_entry.get("attributes", {})
        board_features = board_attrs.get("features", {})

        return cls(
            # Online database has inconsistently cased board types.
            # Since this field is used to match against `targets.json`, we need to ensure consistency is maintained.
            board_type=board_attrs.get("board_type", "").upper(),
            board_name=board_attrs.get("name", ""),
            mbed_os_support=tuple(board_features.get("mbed_os_support", [])),
            mbed_enabled=tuple(board_features.get("mbed_enabled", [])),
            product_code=board_attrs.get("product_code", ""),
            target_type=board_attrs.get("target_type", ""),
            slug=board_attrs.get("slug", ""),
            # TODO: Remove this hard-coded example after demoing.
            # NOTE: Currently we hard-code the build variant for a single board type.
            # This is simply so we can demo the tools to PE. This must be removed and replaced with a proper mechanism
            # for determining the build variant for all platforms. We probably need to add this information to the
            # boards database.
            build_variant=tuple(),
        )

    @classmethod
    def from_offline_board_entry(cls, board_entry: dict) -> "Board":
        """Construct an Board with data from the offline database snapshot."""
        return cls(
            board_type=board_entry.get("board_type", ""),
            board_name=board_entry.get("board_name", ""),
            product_code=board_entry.get("product_code", ""),
            target_type=board_entry.get("target_type", ""),
            slug=board_entry.get("slug", ""),
            mbed_os_support=tuple(board_entry.get("mbed_os_support", [])),
            mbed_enabled=tuple(board_entry.get("mbed_enabled", [])),
            build_variant=tuple(board_entry.get("build_variant", [])),
        )



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Module in charge of CMake file generation."""
import pathlib

import jinja2

from mbed_tools.build._internal.config.config import Config

TEMPLATES_DIRECTORY = pathlib.Path("_internal", "templates")
TEMPLATE_NAME = "mbed_config.tmpl"


def render_mbed_config_cmake_template(config: Config, toolchain_name: str, target_name: str) -> str:
    """Renders the mbed_config jinja template with the target and project config settings.

    Args:
        config: Config object holding information parsed from the mbed config system.
        toolchain_name: Name of the toolchain being used.
        target_name: Name of the target.

    Returns:
        The rendered mbed_config template.
    """
    env = jinja2.Environment(loader=jinja2.PackageLoader("mbed_tools.build", str(TEMPLATES_DIRECTORY)),)
    template = env.get_template(TEMPLATE_NAME)
    config["supported_c_libs"] = [x for x in config["supported_c_libs"][toolchain_name.lower()]]
    context = {"target_name": target_name, "toolchain_name": toolchain_name, **config}
    return template.render(context)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Main cli entry point."""
import logging
import sys

from pkg_resources import get_distribution
from typing import Union, Any

import click

from mbed_tools.lib.logging import set_log_level, MbedToolsHandler

from mbed_tools.cli.configure import configure
from mbed_tools.cli.list_connected_devices import list_connected_devices
from mbed_tools.cli.project_management import new, import_, deploy
from mbed_tools.cli.build import build
from mbed_tools.cli.sterm import sterm

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
LOGGER = logging.getLogger(__name__)


class GroupWithExceptionHandling(click.Group):
    """A click.Group which handles ToolsErrors and logging."""

    def invoke(self, context: click.Context) -> None:
        """Invoke the command group.

        Args:
            context: The current click context.
        """
        # Use the context manager to ensure tools exceptions (expected behaviour) are shown as messages to the user,
        # but all other exceptions (unexpected behaviour) are shown as errors.
        with MbedToolsHandler(LOGGER, context.params["traceback"]) as handler:
            super().invoke(context)

        sys.exit(handler.exit_code)


def print_version(context: click.Context, param: Union[click.Option, click.Parameter], value: bool) -> Any:
    """Print the version of mbed-tools."""
    if not value or context.resilient_parsing:
        return

    version_string = get_distribution("mbed-tools").version
    click.echo(version_string)
    context.exit()


@click.group(cls=GroupWithExceptionHandling, context_settings=CONTEXT_SETTINGS)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Display versions of all Mbed Tools packages.",
)
@click.option(
    "-v",
    "--verbose",
    default=0,
    count=True,
    help="Set the verbosity level, enter multiple times to increase verbosity.",
)
@click.option("-t", "--traceback", is_flag=True, show_default=True, help="Show a traceback when an error is raised.")
def cli(verbose: int, traceback: bool) -> None:
    """Command line tool for interacting with Mbed OS."""
    set_log_level(verbose)


cli.add_command(configure, "configure")
cli.add_command(list_connected_devices, "detect")
cli.add_command(new, "new")
cli.add_command(deploy, "deploy")
cli.add_command(import_, "import")
cli.add_command(build, "compile")
cli.add_command(sterm, "sterm")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Provides the core build system for Mbed OS, which relies on CMake and Ninja as underlying technologies.

The functionality covered in this package includes the following:

- Execution of Mbed Pre-Build stages to determine appropriate configuration for Mbed OS and the build process.
- Invocation of the build process for the command line tools and online build service.
- Export of build instructions to third party command line tools and IDEs.
"""
from mbed_tools.build.build import build_project, generate_build_system
from mbed_tools.build.config import generate_config
from mbed_tools.build.flash import flash_binary



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Shared pytest functionality."""
import platform
import unittest


windows_only = unittest.skipIf(platform.system() != "Windows", reason="Windows required")
darwin_only = unittest.skipIf(platform.system() != "Darwin", reason="Darwin required")
linux_only = unittest.skipIf(platform.system() != "Linux", reason="Linux required")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Helpers for json related functions."""
import json
import logging

from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)


def decode_json_file(path: Path) -> Any:
    """Return the contents of json file."""
    try:
        logger.debug(f"Loading JSON file {path}")
        return json.loads(path.read_text())
    except json.JSONDecodeError:
        logger.error(f"Failed to decode JSON data in the file located at '{path}'")
        raise



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import factory
import pathlib

from typing import List

from mbed_tools.devices.device import Device, ConnectedDevices
from mbed_tools.targets.board import Board


class BoardFactory(factory.Factory):
    class Meta:
        model = Board

    board_type = "TEST"
    board_name = ""
    product_code = ""
    target_type = ""
    slug = ""
    build_variant = []
    mbed_os_support = []
    mbed_enabled = []


class DeviceFactory(factory.Factory):
    class Meta:
        model = Device

    serial_port = ""
    serial_number = ""
    mount_points = [pathlib.Path(".")]
    mbed_board = BoardFactory()


class ConnectedDevicesFactory(factory.Factory):
    class Meta:
        model = ConnectedDevices

    identified_devices: List[Device]
    unidentified_devices: List[Device]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines CandidateDevice model used for device detection."""
from dataclasses import dataclass
from typing import Optional, Tuple, Any, Union, cast
from pathlib import Path


class CandidateDeviceError(ValueError):
    """Base exception raised by a CandidateDevice."""


class USBDescriptorError(CandidateDeviceError):
    """USB descriptor field was not found."""


class FilesystemMountpointError(CandidateDeviceError):
    """Filesystem mount point was not found."""


class DataField:
    """CandidateDevice data attribute descriptor."""

    def __set_name__(self, owner: object, name: str) -> None:
        """Sets the descriptor name, this is called by magic in the owners.__new__ method."""
        self.name = name

    def __get__(self, instance: object, owner: object = None) -> Any:
        """Get the attribute value from the instance."""
        return instance.__dict__.setdefault(self.name, None)


class USBDescriptorHex(DataField):
    """USB descriptor field which cannot be set to an empty value, or an invalid hex value."""

    def __set__(self, instance: object, value: Any) -> None:
        """Prevent setting the descriptor to an empty or invalid hex value."""
        try:
            instance.__dict__[self.name] = _format_hex(value)
        except ValueError:
            raise USBDescriptorError(f"{self.name} cannot be an empty and must be valid hex.")


class USBDescriptorString(DataField):
    """USB descriptor field which cannot be set to an empty value."""

    def __set__(self, instance: object, value: str) -> None:
        """Prevent setting the descriptor to a non-string or empty value."""
        if not value or not isinstance(value, str):
            raise USBDescriptorError(f"{self.name} cannot be an empty field and must be a string.")

        instance.__dict__[self.name] = value


class FilesystemMountpoints(DataField):
    """Data descriptor which must be set to a non-empty list or tuple."""

    def __set__(self, instance: object, value: Union[tuple, list]) -> None:
        """Prevent setting the descriptor to a non-sequence or empty sequence value."""
        if not value or not isinstance(value, (list, tuple)):
            raise FilesystemMountpointError(f"{self.name} must be set to a non-empty list or tuple.")

        instance.__dict__[self.name] = tuple(value)


@dataclass(frozen=True, order=True)
class CandidateDevice:
    """Valid candidate device connected to the host computer.

    We define a CandidateDevice as any USB mass storage device which mounts a filesystem.
    The device may or may not present a serial port.

    Attributes:
        product_id: USB device product ID.
        vendor_id: USB device vendor ID.
        serial_number: USB device serial number.
        mount_points: Filesystem mount points associated with the device.
        serial_port: Serial port associated with the device, this could be None.
    """

    product_id: str = cast(str, USBDescriptorHex())
    vendor_id: str = cast(str, USBDescriptorHex())
    serial_number: str = cast(str, USBDescriptorString())
    mount_points: Tuple[Path, ...] = cast(Tuple[Path], FilesystemMountpoints())
    serial_port: Optional[str] = None


def _format_hex(hex_value: str) -> str:
    """Return hex value with a prefix.

    Accepts hex_value in prefixed (0xff) and unprefixed (ff) formats.
    """
    return hex(int(hex_value, 16))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from textwrap import dedent
from unittest import mock

import pytest

from click.testing import CliRunner

from mbed_tools.cli.project_management import new, import_, deploy


@pytest.fixture
def mock_initialise_project():
    with mock.patch("mbed_tools.cli.project_management.initialise_project") as init_proj:
        yield init_proj


@pytest.fixture
def mock_import_project():
    with mock.patch("mbed_tools.cli.project_management.import_project") as import_proj:
        yield import_proj


@pytest.fixture
def mock_deploy_project():
    with mock.patch("mbed_tools.cli.project_management.deploy_project") as deploy_proj:
        yield deploy_proj


@pytest.fixture
def mock_get_libs():
    with mock.patch("mbed_tools.cli.project_management.get_known_libs") as get_libs:
        yield get_libs


class TestNewCommand:
    def test_calls_new_function_with_correct_args(self, mock_initialise_project):
        CliRunner().invoke(new, ["path", "--create-only"])
        mock_initialise_project.assert_called_once_with(pathlib.Path.cwd() / "path", True)

    def test_echos_mbed_os_message_when_required(self, mock_initialise_project):
        expected = (
            "Creating a new Mbed program at path "
            + "'"
            + str(pathlib.Path.cwd() / "path")
            + "'"
            + ".\nDownloading mbed-os and adding it to the project.\n"
        )

        result = CliRunner().invoke(new, ["path"])

        assert result.output == expected


class TestImportCommand:
    def test_calls_clone_function_with_correct_args(self, mock_import_project):
        CliRunner().invoke(import_, ["url", "dst"])
        mock_import_project.assert_called_once_with("url", pathlib.Path("dst"), True)

    def test_prints_fetched_libs(self, mock_import_project, mock_get_libs):
        mock_methods = {"get_git_reference.return_value": mock.Mock(ref="abcdef", repo_url="https://repo/url")}
        mock_get_libs.return_value = [
            mock.Mock(reference_file=pathlib.Path("test"), source_code_path=pathlib.Path("source"), **mock_methods)
        ]
        expected = """
        Library Name    Repository URL    Path    Git Reference
        --------------  ----------------  ------  ---------------
        test            https://repo/url  source  abcdef
        """

        runner = CliRunner()
        ret = runner.invoke(import_, ["url", "dst"])

        assert dedent(expected) in ret.output

    def test_does_not_print_libs_table_when_skip_resolve_specified(self, mock_import_project, mock_get_libs):
        expected = """
        Library Name    Repository URL    Path    Git Reference
        --------------  ----------------  ------  ---------------
        """

        runner = CliRunner()
        ret = runner.invoke(import_, ["url", "dst", "-s"])

        assert dedent(expected) not in ret.output
        mock_get_libs.assert_not_called()


class TestDeployCommand:
    def test_calls_deploy_function_with_correct_args(self, mock_deploy_project):
        CliRunner().invoke(deploy, ["path", "--force"])
        mock_deploy_project.assert_called_once_with(pathlib.Path("path"), True)

    def test_prints_fetched_libs(self, mock_deploy_project, mock_get_libs):
        mock_methods = {"get_git_reference.return_value": mock.Mock(ref="abcdef", repo_url="https://repo/url")}
        mock_get_libs.return_value = [
            mock.Mock(reference_file=pathlib.Path("test"), source_code_path=pathlib.Path("source"), **mock_methods)
        ]
        expected = """
        Library Name    Repository URL    Path    Git Reference
        --------------  ----------------  ------  ---------------
        test            https://repo/url  source  abcdef
        """

        runner = CliRunner()
        ret = runner.invoke(deploy, ["path", "--force"])

        assert dedent(expected) in ret.output



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock
from mbed_tools.targets.get_target import get_target_by_board_type, get_target_by_name
from mbed_tools.targets.exceptions import TargetError
from mbed_tools.targets._internal.target_attributes import TargetAttributesError


class TestGetTarget(TestCase):
    @mock.patch("mbed_tools.targets.get_target.target_attributes.get_target_attributes")
    def test_get_by_name(self, mock_target_attrs):
        target_name = "Target"
        targets_json_file_path = "targets.json"

        result = get_target_by_name(target_name, targets_json_file_path)

        self.assertEqual(result, mock_target_attrs.return_value)
        mock_target_attrs.assert_called_once_with(targets_json_file_path, target_name)

    @mock.patch("mbed_tools.targets.get_target.target_attributes.get_target_attributes")
    def test_get_by_name_raises_target_error_when_target_json_not_found(self, mock_target_attrs):
        target_name = "Target"
        targets_json_file_path = "not-targets.json"
        mock_target_attrs.side_effect = FileNotFoundError

        with self.assertRaises(TargetError):
            get_target_by_name(target_name, targets_json_file_path)

    @mock.patch("mbed_tools.targets.get_target.target_attributes.get_target_attributes")
    def test_get_by_name_raises_target_error_when_target_attr_collection_fails(self, mock_target_attrs):
        target_name = "Target"
        targets_json_file_path = "targets.json"
        mock_target_attrs.side_effect = TargetAttributesError

        with self.assertRaises(TargetError):
            get_target_by_name(target_name, targets_json_file_path)

    @mock.patch("mbed_tools.targets.get_target.get_target_by_name")
    def test_get_by_board_type(self, mock_get_target_by_name):
        board_type = "Board"
        path_to_mbed_program = "somewhere"

        result = get_target_by_board_type(board_type, path_to_mbed_program)

        self.assertEqual(result, mock_get_target_by_name.return_value)
        mock_get_target_by_name.assert_called_once_with(board_type, path_to_mbed_program)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Project management commands: new, import_, deploy and libs."""
import os
import pathlib

from typing import Any, List

import click
import tabulate

from mbed_tools.project import initialise_project, import_project, get_known_libs, deploy_project
from mbed_tools.project._internal import git_utils


@click.command()
@click.option("--create-only", "-c", is_flag=True, show_default=True, help="Create a program without fetching mbed-os.")
@click.argument("path", type=click.Path(resolve_path=True))
def new(path: str, create_only: bool) -> None:
    """Creates a new Mbed project at the specified path. Downloads mbed-os and adds it to the project.

    PATH: Path to the destination directory for the project. Will be created if it does not exist.
    """
    click.echo(f"Creating a new Mbed program at path '{path}'.")
    if not create_only:
        click.echo("Downloading mbed-os and adding it to the project.")

    initialise_project(pathlib.Path(path), create_only)


@click.command()
@click.argument("url")
@click.argument("path", type=click.Path(), default="")
@click.option(
    "--skip-resolve-libs",
    "-s",
    is_flag=True,
    show_default=True,
    help="Skip resolving program library dependencies after cloning.",
)
def import_(url: str, path: Any, skip_resolve_libs: bool) -> None:
    """Clone an Mbed project and library dependencies.

    URL: The git url of the remote project to clone.

    PATH: Destination path for the clone. If not given the destination path is set to the project name in the cwd.
    """
    click.echo(f"Cloning Mbed program '{url}'")
    if not skip_resolve_libs:
        click.echo("Resolving program library dependencies.")

    if path:
        click.echo(f"Destination path is '{path}'")
        path = pathlib.Path(path)

    dst_path = import_project(url, path, not skip_resolve_libs)
    if not skip_resolve_libs:
        libs = get_known_libs(dst_path)
        _print_dependency_table(libs)


@click.command()
@click.argument("path", type=click.Path(), default=os.getcwd())
@click.option(
    "--force",
    "-f",
    is_flag=True,
    show_default=True,
    help="Forces checkout of all library repositories at specified commit in the .lib file, overwrites local changes.",
)
def deploy(path: str, force: bool) -> None:
    """Checks out Mbed program library dependencies at the revision specified in the ".lib" files.

    Ensures all dependencies are resolved and the versions are synchronised to the version specified in the library
    reference.

    PATH: Path to the Mbed project [default: CWD]
    """
    click.echo("Checking out all libraries to revisions specified in .lib files. Resolving any unresolved libraries.")
    root_path = pathlib.Path(path)
    deploy_project(root_path, force)
    libs = get_known_libs(root_path)
    _print_dependency_table(libs)


def _print_dependency_table(libs: List) -> None:
    click.echo("The following library dependencies were fetched: \n")
    table = []
    for lib in libs:
        table.append(
            [
                lib.reference_file.stem,
                lib.get_git_reference().repo_url,
                lib.source_code_path,
                git_utils.get_default_branch(git_utils.get_repo(lib.source_code_path))
                if not lib.get_git_reference().ref
                else lib.get_git_reference().ref,
            ]
        )

    headers = ("Library Name", "Repository URL", "Path", "Git Reference")
    click.echo(tabulate.tabulate(table, headers=headers))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Internal helper to retrieve target information from the online database."""

import pathlib
from http import HTTPStatus
import json
from json.decoder import JSONDecodeError
import logging
from typing import List, Optional, Dict, Any

import requests

from mbed_tools.targets._internal.exceptions import ResponseJSONError, BoardAPIError

from mbed_tools.targets.env import env


INTERNAL_PACKAGE_DIR = pathlib.Path(__file__).parent
SNAPSHOT_FILENAME = "board_database_snapshot.json"

logger = logging.getLogger(__name__)


def get_board_database_path() -> pathlib.Path:
    """Return the path to the offline board database."""
    return pathlib.Path(INTERNAL_PACKAGE_DIR, "data", SNAPSHOT_FILENAME)


_BOARD_API = "https://os.mbed.com/api/v4/targets"


def get_offline_board_data() -> Any:
    """Loads board data from JSON stored in offline snapshot.

    Returns:
        The board database as retrieved from the local database snapshot.

    Raises:
        ResponseJSONError: error decoding the local database JSON.
    """
    boards_snapshot_path = get_board_database_path()
    try:
        return json.loads(boards_snapshot_path.read_text())
    except JSONDecodeError as json_err:
        raise ResponseJSONError(f"Invalid JSON received from '{boards_snapshot_path}'.") from json_err


def get_online_board_data() -> List[dict]:
    """Retrieves board data from the online API.

    Returns:
        The board database as retrieved from the boards API

    Raises:
        ResponseJSONError: error decoding the response JSON.
        BoardAPIError: error retrieving data from the board API.
    """
    board_data: List[dict] = [{}]
    response = _get_request()
    if response.status_code != HTTPStatus.OK:
        warning_msg = _response_error_code_to_str(response)
        logger.warning(warning_msg)
        logger.debug(f"Response received from API:\n{response.text}")
        raise BoardAPIError(warning_msg)

    try:
        json_data = response.json()
    except JSONDecodeError as json_err:
        warning_msg = f"Invalid JSON received from '{_BOARD_API}'."
        logger.warning(warning_msg)
        logger.debug(f"Response received from API:\n{response.text}")
        raise ResponseJSONError(warning_msg) from json_err

    try:
        board_data = json_data["data"]
    except KeyError as key_err:
        warning_msg = f"JSON received from '{_BOARD_API}' is missing the 'data' field."
        logger.warning(warning_msg)
        keys_found = ", ".join(json_data.keys())
        logger.debug(f"Fields found in JSON Response: {keys_found}")
        raise ResponseJSONError(warning_msg) from key_err

    return board_data


def _response_error_code_to_str(response: requests.Response) -> str:
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        return (
            f"Authentication failed for '{_BOARD_API}'. Please check that the environment variable "
            f"'MBED_API_AUTH_TOKEN' is correctly configured with a private access token."
        )
    else:
        return f"An HTTP {response.status_code} was received from '{_BOARD_API}'."


def _get_request() -> requests.Response:
    """Make a GET request to the API, ensuring the correct headers are set."""
    header: Optional[Dict[str, str]] = None
    mbed_api_auth_token = env.MBED_API_AUTH_TOKEN
    if mbed_api_auth_token:
        header = {"Authorization": f"Bearer {mbed_api_auth_token}"}

    try:
        return requests.get(_BOARD_API, headers=header)
    except requests.exceptions.ConnectionError as connection_error:
        if isinstance(connection_error, requests.exceptions.SSLError):
            logger.warning("Unable to verify an SSL certificate with requests.")
        elif isinstance(connection_error, requests.exceptions.ProxyError):
            logger.warning("Failed to connect to proxy. Please check your proxy configuration.")

        logger.warning("Unable to connect to the online database. Please check your internet connection.")
        raise BoardAPIError("Failed to connect to the online database.") from connection_error



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Data aggregation about a disk on Windows.

On Windows, information about disk drive is scattered around Physical disks,
Partitions and Logical Drives.
This file tries to reconcile all these pieces of information so that it is presented
as a single object: AggregatedDiskData.
"""
from typing import List, Optional, Callable
from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor
from mbed_tools.devices._internal.windows.component_descriptor_utils import retain_value_or_default
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID
from mbed_tools.devices._internal.windows.disk_drive import DiskDrive
from mbed_tools.devices._internal.windows.disk_partition import DiskPartition
from mbed_tools.devices._internal.windows.disk_partition_logical_disk_relationships import (
    DiskPartitionLogicalDiskRelationship,
)
from mbed_tools.devices._internal.windows.logical_disk import LogicalDisk
from mbed_tools.devices._internal.windows.volume_set import VolumeInformation, get_volume_information
from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader, ComponentsLoader


class AggregatedDiskDataDefinition(NamedTuple):
    """Data aggregation with regards to a disk."""

    uid: WindowsUID  # see WindowsUID
    label: str  # e.g. C:
    description: str  # e.g. Removal Disk
    free_space: int
    size: int
    partition_name: str  # e.g. Disk #0, Partition #2
    partition_type: str  # 16-bit FAT
    volume_information: VolumeInformation
    caption: str  # e.g. SAMSUNG MZNLN512HMJP-000L7
    physical_disk_name: str
    model: str
    interface_type: str  # e.g. IDE
    media_type: str  # e.g. Fixed hard disk media
    manufacturer: str
    serial_number: str
    status: str
    pnp_device_id: str


class AggregatedDiskData(ComponentDescriptor):
    """Disk information based on lots of different sources."""

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(AggregatedDiskDataDefinition, win32_class_name="DiskDataAggregation")

    @property
    def component_id(self) -> str:
        """Returns the ID field."""
        return cast(str, self.get("label"))


class DiskDataAggregator:
    """Aggregator for any scattered data related to disks."""

    def __init__(
        self,
        physical_disks: dict,
        partition_disks: dict,
        logical_partition_relationships: dict,
        lookup_volume_information: Callable[[LogicalDisk], VolumeInformation],
    ) -> None:
        """Initialiser."""
        self._physical_disks = physical_disks
        self._partition_disks = partition_disks
        self._logical_partition_relationships = logical_partition_relationships
        self._lookup_volume_information = lookup_volume_information

    def _get_corresponding_partition(self, logical_disk: LogicalDisk) -> DiskPartition:
        """Determines the partition corresponding to a logical disk."""
        # Determines the partition ID
        partition_id = self.logical_disk_partition_relationships.get(logical_disk.component_id)
        return self._partition_disks.get(partition_id, DiskPartition()) if partition_id else DiskPartition()

    @property
    def physical_disks(self) -> dict:
        """Gets local cache of physical disks data."""
        return self._physical_disks

    @property
    def partition_disks(self) -> dict:
        """Gets local cache of disk partition data."""
        return self._partition_disks

    @property
    def logical_disk_partition_relationships(self) -> dict:
        """Gets local cache of relationships between logical disk and disk partitions."""
        return self._logical_partition_relationships

    def aggregate(self, logical_disk: LogicalDisk) -> AggregatedDiskData:
        """Aggregates data about a disk from different sources."""
        corresponding_partition = self._get_corresponding_partition(logical_disk)
        corresponding_volume_information = self._lookup_volume_information(logical_disk)
        # Determines which physical disk the partition is on
        # See https://superuser.com/questions/1147218/on-which-physical-drive-is-this-logical-drive
        corresponding_physical = self._physical_disks.get(corresponding_partition.get("DiskIndex"), DiskDrive())
        aggregatedData = AggregatedDiskData()
        aggregatedData.set_data_values(
            dict(
                uid=corresponding_physical.uid,
                label=logical_disk.component_id,
                description=logical_disk.get("Description"),
                free_space=logical_disk.get("FreeSpace"),
                size=logical_disk.get("Size"),
                partition_name=corresponding_partition.component_id,
                partition_type=corresponding_partition.get("Type"),
                volume_information=corresponding_volume_information,
                caption=corresponding_physical.get("Caption"),
                physical_disk_name=corresponding_physical.get("DeviceID"),
                model=corresponding_physical.get("Model"),
                interface_type=corresponding_physical.get("InterfaceType"),
                media_type=corresponding_physical.get("MediaType"),
                manufacturer=corresponding_physical.get("Manufacturer"),
                serial_number=retain_value_or_default(corresponding_physical.get("SerialNumber")),
                status=corresponding_physical.get("Status"),
                pnp_device_id=corresponding_physical.get("PNPDeviceID"),
            )
        )
        return aggregatedData


class WindowsDiskDataAggregator(DiskDataAggregator):
    """Disk Data aggregator for Windows."""

    def __init__(self, data_loader: SystemDataLoader) -> None:
        """Initialiser."""
        super().__init__(
            physical_disks={
                d.Index: d  # type: ignore
                for d in ComponentsLoader(data_loader, DiskDrive).element_generator()
            },
            partition_disks={
                p.component_id: p for p in ComponentsLoader(data_loader, DiskPartition).element_generator()
            },
            logical_partition_relationships={
                r.logical_disk_id: r.disk_partition_id  # type: ignore
                for r in ComponentsLoader(data_loader, DiskPartitionLogicalDiskRelationship).element_generator()
            },
            lookup_volume_information=lambda logical_disk: get_volume_information(logical_disk.component_id),
        )


class SystemDiskInformation:
    """All information about disks on the current system."""

    def __init__(self, data_loader: SystemDataLoader) -> None:
        """Initialiser."""
        self._disk_data_by_serial_number: Optional[dict] = None
        self._disk_data_by_label: Optional[dict] = None
        self._data_loader = data_loader

    def _load_data(self) -> None:
        aggregator = WindowsDiskDataAggregator(self._data_loader)
        disk_data_by_serialnumber: dict = dict()  # The type is enforced so that mypy is happy.
        disk_data_by_label = dict()
        for ld in ComponentsLoader(self._data_loader, LogicalDisk).element_generator():
            aggregation = aggregator.aggregate(cast(LogicalDisk, ld))
            key = aggregation.get("uid").presumed_serial_number
            disk_data_list = disk_data_by_serialnumber.get(key, list())
            disk_data_list.append(aggregation)
            disk_data_by_serialnumber[key] = disk_data_list
            disk_data_by_label[aggregation.get("label")] = aggregation
        self._disk_data_by_serial_number = disk_data_by_serialnumber
        self._disk_data_by_label = disk_data_by_label

    @property
    def disk_data_by_serial_number(self) -> dict:
        """Gets system's disk data by serial number."""
        if not self._disk_data_by_serial_number:
            self._load_data()
        return self._disk_data_by_serial_number if self._disk_data_by_serial_number else dict()

    @property
    def disk_data_by_label(self) -> dict:
        """Gets system's disk data by label."""
        if not self._disk_data_by_label:
            self._load_data()
        return self._disk_data_by_label if self._disk_data_by_label else dict()

    def get_disk_information(self, uid: WindowsUID) -> List[AggregatedDiskData]:
        """Gets all disk information for a given UID."""
        return self.disk_data_by_serial_number.get(uid.presumed_serial_number, list())

    def get_disk_information_by_label(self, label: str) -> AggregatedDiskData:
        """Gets all disk information for a given label."""
        return self.disk_data_by_label.get(
            label.upper(), self.disk_data_by_label.get(label.lower(), AggregatedDiskData())
        )



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Helpers for logging errors according to severity of the exception."""
from typing import Type, Optional, cast
from types import TracebackType
import logging
from mbed_tools.lib.exceptions import ToolsError

LOGGING_FORMAT = "%(levelname)s: %(message)s"

VERBOSITY_HELP = {
    logging.CRITICAL: "-v",
    logging.ERROR: "-v",
    logging.WARNING: "-vv",
    logging.INFO: "-vvv",
    logging.DEBUG: "--traceback",
}


def _exception_message(err: BaseException, log_level: int, traceback: bool) -> str:
    """Generate a user facing message with help on how to get more information from the logs."""
    error_msg = str(err)
    if log_level != logging.DEBUG or not traceback:
        cli_option = VERBOSITY_HELP.get(log_level, "-v")
        error_msg += f"\n\nMore information may be available by using the command line option '{cli_option}'."
    return error_msg


class MbedToolsHandler:
    """Context Manager to catch Mbed Tools exceptions and generate a helpful user facing message."""

    def __init__(self, logger: logging.Logger, traceback: bool = False):
        """Keep track of the logger to use and whether or not a traceback should be generated."""
        self._logger = logger
        self._traceback = traceback
        self.exit_code = 0

    def __enter__(self) -> "MbedToolsHandler":
        """Return the Context Manager."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> bool:
        """Handle any raised exceptions, suppressing Tools errors and generating an error message instead."""
        if exc_type and issubclass(exc_type, ToolsError):
            error_msg = _exception_message(cast(BaseException, exc_value), logging.root.level, self._traceback)
            self._logger.error(error_msg, exc_info=self._traceback)
            # Do not propagate exceptions derived from ToolsError
            self.exit_code = 1
            return True

        # Propagate all other exceptions
        return False


def log_exception(logger: logging.Logger, exception: Exception, show_traceback: bool = False) -> None:
    """Logs an exception in both normal and verbose forms.

    Args:
        logger: logger
        exception: exception to log
        show_traceback: show the full traceback.
    """
    logger.error(exception, exc_info=show_traceback)


def set_log_level(verbose_count: int) -> None:
    """Sets the log level.

    Args:
        verbose_count: number of `-v` flags used
    """
    if verbose_count > 2:
        log_level = logging.DEBUG
    elif verbose_count == 2:
        log_level = logging.INFO
    elif verbose_count == 1:
        log_level = logging.WARNING
    else:
        log_level = logging.ERROR
    logging.basicConfig(level=log_level, format=LOGGING_FORMAT)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pytest

from mbed_tools.build._internal.cmake_file import render_mbed_config_cmake_template
from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.config.source import ConfigSetting, prepare


TOOLCHAIN_NAME = "gcc"


@pytest.fixture()
def fake_target():
    return {
        "labels": ["foo"],
        "extra_labels": ["morefoo"],
        "features": ["bar"],
        "components": ["baz"],
        "macros": ["macbaz"],
        "device_has": ["stuff"],
        "c_lib": ["c_lib"],
        "core": ["core"],
        "printf_lib": ["printf_lib"],
        "supported_form_factors": ["arduino"],
        "supported_c_libs": {TOOLCHAIN_NAME: ["ginormous"]},
        "supported_application_profiles": ["full", "bare-metal"],
    }


class TestRendersCMakeListsFile:
    def test_returns_rendered_content(self, fake_target):
        config = Config(prepare(fake_target))
        result = render_mbed_config_cmake_template(config, TOOLCHAIN_NAME, "target_name")

        for label in fake_target["labels"] + fake_target["extra_labels"]:
            assert label in result

        for macro in fake_target["features"] + fake_target["components"] + [TOOLCHAIN_NAME]:
            assert macro in result

        for toolchain in fake_target["supported_c_libs"]:
            assert toolchain in result
            for supported_c_libs in toolchain:
                assert supported_c_libs in result

        for supported_application_profiles in fake_target["supported_application_profiles"]:
            assert supported_application_profiles in result

    def test_returns_quoted_content(self, fake_target):
        config = Config(prepare(fake_target))

        # Add an option whose value contains quotes to the config.
        config["config"] = [
            ConfigSetting(
                name="mqtt-host", namespace="iotc", help_text="", value='{"mqtt.2030.ltsapis.goog", IOTC_MQTT_PORT}',
            )
        ]

        result = render_mbed_config_cmake_template(config, TOOLCHAIN_NAME, "target_name")
        assert '"-DMBED_CONF_IOTC_MQTT_HOST={\\"mqtt.2030.ltsapis.goog\\", IOTC_MQTT_PORT}"' in result



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Objects representing Mbed program and library data."""
import json
import logging

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from mbed_tools.project._internal.render_templates import (
    render_cmakelists_template,
    render_main_cpp_template,
    render_gitignore_template,
)

logger = logging.getLogger(__name__)

# Mbed program file names and constants.
APP_CONFIG_FILE_NAME = "mbed_app.json"
BUILD_DIR = "cmake_build"
CMAKELISTS_FILE_NAME = "CMakeLists.txt"
MAIN_CPP_FILE_NAME = "main.cpp"
MBED_OS_REFERENCE_FILE_NAME = "mbed-os.lib"
MBED_OS_DIR_NAME = "mbed-os"
TARGETS_JSON_FILE_PATH = Path("targets", "targets.json")
CUSTOM_TARGETS_JSON_FILE_NAME = "custom_targets.json"

# Information written to mbed-os.lib
MBED_OS_REFERENCE_URL = "https://github.com/ARMmbed/mbed-os"
MBED_OS_REFERENCE_ID = "master"

# For some reason Mbed OS expects the default mbed_app.json to contain some target_overrides
# for the K64F target. TODO: Find out why this wouldn't be defined in targets.json.
DEFAULT_APP_CONFIG = {"target_overrides": {"K64F": {"platform.stdio-baud-rate": 9600}}}


@dataclass
class MbedProgramFiles:
    """Files defining an MbedProgram.

    This object holds paths to the various files which define an MbedProgram.

    MbedPrograms must contain an mbed-os.lib reference file, defining Mbed OS as a program dependency. A program can
    optionally include an mbed_app.json file which defines application level config.

    Attributes:
        app_config_file: Path to mbed_app.json file. This can be `None` if the program doesn't set any custom config.
        mbed_os_ref: Library reference file for MbedOS. All programs require this file.
        cmakelists_file: A top-level CMakeLists.txt containing build definitions for the application.
        cmake_build_dir: The CMake build tree.
    """

    app_config_file: Optional[Path]
    mbed_os_ref: Path
    cmakelists_file: Path
    cmake_build_dir: Path
    custom_targets_json: Path

    @classmethod
    def from_new(cls, root_path: Path) -> "MbedProgramFiles":
        """Create MbedProgramFiles from a new directory.

        A "new directory" in this context means it doesn't already contain an Mbed program.

        Args:
            root_path: The directory in which to create the program data files.

        Raises:
            ValueError: A program .mbed or mbed-os.lib file already exists at this path.
        """
        app_config = root_path / APP_CONFIG_FILE_NAME
        mbed_os_ref = root_path / MBED_OS_REFERENCE_FILE_NAME
        cmakelists_file = root_path / CMAKELISTS_FILE_NAME
        main_cpp = root_path / MAIN_CPP_FILE_NAME
        gitignore = root_path / ".gitignore"
        cmake_build_dir = root_path / BUILD_DIR
        custom_targets_json = root_path / CUSTOM_TARGETS_JSON_FILE_NAME

        if mbed_os_ref.exists():
            raise ValueError(f"Program already exists at path {root_path}.")

        app_config.write_text(json.dumps(DEFAULT_APP_CONFIG, indent=4))
        mbed_os_ref.write_text(f"{MBED_OS_REFERENCE_URL}#{MBED_OS_REFERENCE_ID}")
        render_cmakelists_template(cmakelists_file, root_path.stem)
        render_main_cpp_template(main_cpp)
        render_gitignore_template(gitignore)
        return cls(
            app_config_file=app_config,
            mbed_os_ref=mbed_os_ref,
            cmakelists_file=cmakelists_file,
            cmake_build_dir=cmake_build_dir,
            custom_targets_json=custom_targets_json,
        )

    @classmethod
    def from_existing(cls, root_path: Path, build_subdir: Path) -> "MbedProgramFiles":
        """Create MbedProgramFiles from a directory containing an existing program.

        Args:
            root_path: The path containing the MbedProgramFiles.
            build_subdir: The subdirectory of BUILD_DIR to use for CMake build.
        """
        app_config: Optional[Path]
        app_config = root_path / APP_CONFIG_FILE_NAME
        if not app_config.exists():
            logger.info("This program does not contain an mbed_app.json config file.")
            app_config = None

        custom_targets_json = root_path / CUSTOM_TARGETS_JSON_FILE_NAME
        mbed_os_file = root_path / MBED_OS_REFERENCE_FILE_NAME

        cmakelists_file = root_path / CMAKELISTS_FILE_NAME
        if not cmakelists_file.exists():
            logger.warning("No CMakeLists.txt found in the program root.")
        cmake_build_dir = root_path / BUILD_DIR / build_subdir

        return cls(
            app_config_file=app_config,
            mbed_os_ref=mbed_os_file,
            cmakelists_file=cmakelists_file,
            cmake_build_dir=cmake_build_dir,
            custom_targets_json=custom_targets_json,
        )


@dataclass
class MbedOS:
    """Metadata associated with a copy of MbedOS.

    This object holds information about MbedOS used by MbedProgram.

    Attributes:
        root: The root path of the MbedOS source tree.
        targets_json_file: Path to a targets.json file, which contains target data specific to MbedOS revision.
    """

    root: Path
    targets_json_file: Path

    @classmethod
    def from_existing(cls, root_path: Path, check_root_path_exists: bool = True) -> "MbedOS":
        """Create MbedOS from a directory containing an existing MbedOS installation."""
        targets_json_file = root_path / TARGETS_JSON_FILE_PATH

        if check_root_path_exists and not root_path.exists():
            raise ValueError("The mbed-os directory does not exist.")

        if root_path.exists() and not targets_json_file.exists():
            raise ValueError("This MbedOS copy does not contain a targets.json file.")

        return cls(root=root_path, targets_json_file=targets_json_file)

    @classmethod
    def from_new(cls, root_path: Path) -> "MbedOS":
        """Create MbedOS from an empty or new directory."""
        return cls(root=root_path, targets_json_file=root_path / TARGETS_JSON_FILE_PATH)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets.get_board`."""
from unittest import mock, TestCase

from mbed_tools.targets._internal.exceptions import BoardAPIError

# Import from top level as this is the expected interface for users
from mbed_tools.targets import get_board_by_online_id, get_board_by_product_code
from mbed_tools.targets.get_board import (
    _DatabaseMode,
    _get_database_mode,
    get_board,
)
from mbed_tools.targets.env import env
from mbed_tools.targets.exceptions import UnknownBoard, UnsupportedMode
from tests.targets.factories import make_board


@mock.patch("mbed_tools.targets.get_board.Boards", autospec=True)
@mock.patch("mbed_tools.targets.get_board.env", spec_set=env)
class TestGetBoard(TestCase):
    def test_online_mode(self, env, mocked_boards):
        env.MBED_DATABASE_MODE = "ONLINE"
        fn = mock.Mock()

        subject = get_board(fn)

        self.assertEqual(subject, mocked_boards.from_online_database().get_board.return_value)
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)

    def test_offline_mode(self, env, mocked_boards):
        env.MBED_DATABASE_MODE = "OFFLINE"
        fn = mock.Mock()

        subject = get_board(fn)

        self.assertEqual(subject, mocked_boards.from_offline_database().get_board.return_value)
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_calls_offline_boards_first(self, env, mocked_boards):
        env.MBED_DATABASE_MODE = "AUTO"
        fn = mock.Mock()

        subject = get_board(fn)

        self.assertEqual(subject, mocked_boards.from_offline_database().get_board.return_value)
        mocked_boards.from_online_database().get_board.assert_not_called()
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_falls_back_to_online_database_when_board_not_found(self, env, mocked_boards):
        env.MBED_DATABASE_MODE = "AUTO"
        mocked_boards.from_offline_database().get_board.side_effect = UnknownBoard
        fn = mock.Mock()

        subject = get_board(fn)

        self.assertEqual(subject, mocked_boards.from_online_database().get_board.return_value)
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_raises_when_board_not_found_offline_with_no_network(self, env, mocked_boards):
        env.MBED_DATABASE_MODE = "AUTO"
        mocked_boards.from_offline_database().get_board.side_effect = UnknownBoard
        mocked_boards.from_online_database().get_board.side_effect = BoardAPIError
        fn = mock.Mock()

        with self.assertRaises(UnknownBoard):
            get_board(fn)
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)


class TestGetBoardByProductCode(TestCase):
    @mock.patch("mbed_tools.targets.get_board.get_board")
    def test_matches_boards_by_product_code(self, mock_get_board):
        product_code = "swag"

        self.assertEqual(get_board_by_product_code(product_code), mock_get_board.return_value)

        # Test callable matches correct boards
        fn = mock_get_board.call_args[0][0]

        matching_board = make_board(product_code=product_code)
        not_matching_board = make_board(product_code="whatever")

        self.assertTrue(fn(matching_board))
        self.assertFalse(fn(not_matching_board))


class TestGetBoardByOnlineId(TestCase):
    @mock.patch("mbed_tools.targets.get_board.get_board")
    def test_matches_boards_by_online_id(self, mock_get_board):
        target_type = "platform"

        self.assertEqual(get_board_by_online_id(slug="slug", target_type=target_type), mock_get_board.return_value)

        # Test callable matches correct boards
        fn = mock_get_board.call_args[0][0]

        matching_board_1 = make_board(target_type=target_type, slug="slug")
        matching_board_2 = make_board(target_type=target_type, slug="SlUg")
        not_matching_board = make_board(target_type=target_type, slug="whatever")

        self.assertTrue(fn(matching_board_1))
        self.assertTrue(fn(matching_board_2))
        self.assertFalse(fn(not_matching_board))


@mock.patch("mbed_tools.targets.get_board.env", spec_set=env)
class TestGetDatabaseMode(TestCase):
    def test_returns_configured_database_mode(self, env):
        env.MBED_DATABASE_MODE = "OFFLINE"
        self.assertEqual(_get_database_mode(), _DatabaseMode.OFFLINE)

    def test_raises_when_configuration_is_not_supported(self, env):
        env.MBED_DATABASE_MODE = "NOT_VALID"
        with self.assertRaises(UnsupportedMode):
            _get_database_mode()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Exceptions used internally by the mbed-targets package."""

from mbed_tools.targets.exceptions import BoardDatabaseError, MbedTargetsError


class BoardAPIError(BoardDatabaseError):
    """API request failed."""


class ResponseJSONError(BoardDatabaseError):
    """HTTP response JSON parsing failed."""


class TargetsJsonConfigurationError(MbedTargetsError):
    """The target definition is invalid."""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Synchronise the offline target database with the online database.

This utility performs the following actions:
* Downloads the latest online target database
* Saves the database to a local file in the mbed-tools repository
* Writes a news file detailing any added, removed or modified boards
"""

import argparse
import logging
import sys

from dataclasses import dataclass
from pathlib import Path

from mbed_tools_ci_scripts.create_news_file import create_news_file, NewsType
from mbed_tools.lib.exceptions import ToolsError
from mbed_tools.lib.logging import log_exception, set_log_level

from mbed_tools.targets._internal.board_database import get_board_database_path
from mbed_tools.targets.boards import Boards

logger = logging.getLogger()

BOARD_DATABASE_PATH = get_board_database_path()


@dataclass(frozen=True)
class DatabaseComparisonResult:
    """Result of the database comparison."""

    boards_added: set
    boards_removed: set
    boards_modified: set


def save_board_database(board_database_text: str, output_file_path: Path) -> None:
    """Save a snapshot of the board database to a local file.

    Args:
        board_database_text: json formatted text containing the board data returned from the online database
        output_file_path: the path to the output file
    """
    output_file_path.parent.mkdir(exist_ok=True)
    output_file_path.write_text(board_database_text)


def compare_databases(offline_boards: Boards, online_boards: Boards) -> DatabaseComparisonResult:
    """Compare offline and online board databases."""
    added = online_boards - offline_boards
    removed = offline_boards - online_boards
    added_board_names = set(b.board_name for b in added)
    removed_board_names = set(b.board_name for b in removed)
    modified_board_names = added_board_names & removed_board_names
    added_board_names -= modified_board_names
    removed_board_names -= modified_board_names
    return DatabaseComparisonResult(added_board_names, removed_board_names, modified_board_names)


def create_news_item_text_from_boards(prefix: str, board_names: set) -> str:
    """Create a news item string from the list of boards."""
    return f"{prefix} {', '.join(sorted(board_names))}.\n"


def create_news_file_text_from_result(result: DatabaseComparisonResult) -> str:
    """Creates and writes a news file from the result of the database update.

    Args:
        result: Result of the database update
    """
    news_item_text = ""
    if result.boards_added:
        news_item_text = create_news_item_text_from_boards("Targets added:", result.boards_added)

    if result.boards_removed:
        news_item_text = create_news_item_text_from_boards(f"{news_item_text}Targets removed:", result.boards_removed)

    if result.boards_modified:
        news_item_text = create_news_item_text_from_boards(f"{news_item_text}Targets modified:", result.boards_modified)

    return news_item_text


def parse_args() -> argparse.Namespace:
    """Parse the command line."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", "-v", action="count", default=0)
    return parser.parse_args()


def main(args: argparse.Namespace) -> int:
    """Main entry point."""
    set_log_level(args.verbose)
    try:
        online_boards = Boards.from_online_database()
        offline_boards = Boards.from_offline_database()
        result = compare_databases(offline_boards, online_boards)
        if not (result.boards_added or result.boards_removed or result.boards_modified):
            logger.info("No changes to commit. Exiting.")
            return 0

        news_file_text = create_news_file_text_from_result(result)
        create_news_file(news_file_text, NewsType.feature)
        save_board_database(online_boards.json_dump(), BOARD_DATABASE_PATH)
        return 0
    except ToolsError as tools_error:
        log_exception(logger, tools_error)
        return 1


if __name__ == "__main__":
    sys.exit(main(parse_args()))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to list all Mbed enabled devices connected to the host computer."""
import click
import json
from operator import attrgetter
from typing import Iterable, List, Optional, Tuple
from tabulate import tabulate

from mbed_tools.devices import get_connected_devices, Device
from mbed_tools.targets import Board


@click.command()
@click.option(
    "--format", type=click.Choice(["table", "json"]), default="table", show_default=True, help="Set output format."
)
@click.option(
    "--show-all",
    "-a",
    is_flag=True,
    default=False,
    help="Show all connected devices, even those which are not Mbed Boards.",
)
def list_connected_devices(format: str, show_all: bool) -> None:
    """Prints connected devices."""
    connected_devices = get_connected_devices()

    if show_all:
        devices = _sort_devices(connected_devices.identified_devices + connected_devices.unidentified_devices)
    else:
        devices = _sort_devices(connected_devices.identified_devices)

    output_builders = {
        "table": _build_tabular_output,
        "json": _build_json_output,
    }
    if devices:
        output = output_builders[format](devices)
        click.echo(output)
    else:
        click.echo("No connected Mbed devices found.")


def _sort_devices(devices: Iterable[Device]) -> Iterable[Device]:
    """Sort devices by board name and then serial number (in case there are multiple boards with the same name)."""
    return sorted(devices, key=attrgetter("mbed_board.board_name", "serial_number"))


def _get_devices_ids(devices: Iterable[Device]) -> List[Tuple[Optional[int], Device]]:
    """Create tuple of ID and Device for each Device. ID is None when only one Device exists with a given board name."""
    devices_ids: List[Tuple[Optional[int], Device]] = []
    n = 0
    for device in devices:
        board_name = device.mbed_board.board_name
        if len([dev for dev in devices if dev.mbed_board.board_name == board_name]) > 1:
            devices_ids.append((n, device))
            n += 1
        else:
            devices_ids.append((None, device))
            n = 0
    return devices_ids


def _build_tabular_output(devices: Iterable[Device]) -> str:
    headers = ["Board name", "Serial number", "Serial port", "Mount point(s)", "Build target(s)"]
    devices_data = []
    for id, device in _get_devices_ids(devices):
        devices_data.append(
            [
                device.mbed_board.board_name or "<unknown>",
                device.serial_number,
                device.serial_port or "<unknown>",
                "\n".join(str(mount_point) for mount_point in device.mount_points),
                "\n".join(_get_build_targets(device.mbed_board, id)),
            ]
        )
    return tabulate(devices_data, headers=headers)


def _build_json_output(devices: Iterable[Device]) -> str:
    devices_data = []
    for id, device in _get_devices_ids(devices):
        board = device.mbed_board
        devices_data.append(
            {
                "serial_number": device.serial_number,
                "serial_port": device.serial_port,
                "mount_points": [str(m) for m in device.mount_points],
                "mbed_board": {
                    "product_code": board.product_code,
                    "board_type": board.board_type,
                    "board_name": board.board_name,
                    "mbed_os_support": board.mbed_os_support,
                    "mbed_enabled": board.mbed_enabled,
                    "build_targets": _get_build_targets(board, id),
                },
            }
        )
    return json.dumps(devices_data, indent=4)


def _get_build_targets(board: Board, identifier: Optional[int]) -> List[str]:
    if identifier is not None:
        return [f"{board.board_type}_{variant}[{identifier}]" for variant in board.build_variant] + [
            f"{board.board_type}[{identifier}]"
        ]
    else:
        return [f"{board.board_type}_{variant}" for variant in board.build_variant] + [board.board_type]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Build configuration representation."""
import logging

from collections import UserDict
from typing import Any, Iterable, Hashable, Callable, List

from mbed_tools.build._internal.config.source import Override, ConfigSetting

logger = logging.getLogger(__name__)


class Config(UserDict):
    """Mapping of config settings.

    This object understands how to populate the different 'config sections' which all have different rules for how the
    settings are collected.
    Applies overrides, appends macros and updates config settings.
    """

    def __setitem__(self, key: Hashable, item: Any) -> None:
        """Set an item based on its key."""
        if key == CONFIG_SECTION:
            self._update_config_section(item)
        elif key == OVERRIDES_SECTION:
            self._handle_overrides(item)
        elif key == MACROS_SECTION:
            self.data[MACROS_SECTION] = self.data.get(MACROS_SECTION, set()) | item
        else:
            super().__setitem__(key, item)

    def _handle_overrides(self, overrides: Iterable[Override]) -> None:
        for override in overrides:
            logger.debug("Applying override '%s.%s'='%s'", override.namespace, override.name, repr(override.value))
            if override.name == "requires":
                self.data["requires"] = self.data.get("requires", set()) | override.value
                continue

            if override.name in self.data:
                _apply_override(self.data, override)
                continue

            try:
                setting = self._find_first_config_setting(
                    lambda x: x.name == override.name and x.namespace == override.namespace
                )
                setting.value = override.value
            except ValueError:
                logger.warning(
                    f"You are attempting to override an undefined config parameter "
                    f"`{override.namespace}.{override.name}`.\n"
                    "It is an error to override an undefined configuration parameter. "
                    "Please check your target_overrides are correct.\n"
                    f"The parameter `{override.namespace}.{override.name}` will not be added to the configuration."
                )

    def _update_config_section(self, config_settings: List[ConfigSetting]) -> None:
        for setting in config_settings:
            logger.debug("Adding config setting: '%s.%s'", setting.namespace, setting.name)
            if setting in self.data.get(CONFIG_SECTION, []):
                raise ValueError(
                    f"Setting {setting.namespace}.{setting.name} already defined. You cannot duplicate config settings!"
                )

        self.data[CONFIG_SECTION] = self.data.get(CONFIG_SECTION, []) + config_settings

    def _find_first_config_setting(self, predicate: Callable) -> Any:
        """Find first config setting based on `predicate`.

        `predicate` is a callable that gets a config setting passed in as an argument. This callable must define the
        condition for identifying a config setting.

        Example:
            The following call will find the first ConfigSetting whose name is "foo":
            `config._find_config_setting(lambda x: x.name == "foo")`

        Args:
            predicate: A callable that returns True when the desired setting is found.
        """
        for elem in self.data.get(CONFIG_SECTION, []):
            if predicate(elem):
                return elem

        raise ValueError("Could not find element.")


CONFIG_SECTION = "config"
MACROS_SECTION = "macros"
OVERRIDES_SECTION = "overrides"


def _apply_override(data: dict, override: Override) -> None:
    if override.modifier == "add":
        data[override.name] |= override.value
    elif override.modifier == "remove":
        data[override.name] -= override.value
    else:
        data[override.name] = override.value



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Public exceptions raised by the package."""
from mbed_tools.lib.exceptions import ToolsError


class MbedBuildError(ToolsError):
    """Base public exception for the mbed-build package."""


class InvalidExportOutputDirectory(MbedBuildError):
    """It is not possible to export to the provided output directory."""


class BinaryFileNotFoundError(MbedBuildError):
    """The binary file (.bin/.hex) cannot be found in cmake_build directory."""


class DeviceNotFoundError(MbedBuildError):
    """The requested device is not connected to your system."""


class InvalidConfigOverride(MbedBuildError):
    """A given config setting was invalid."""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a device detector for Linux."""
import logging
from pathlib import Path
from typing import Tuple, List, Optional, cast

import psutil
import pyudev

from mbed_tools.devices._internal.base_detector import DeviceDetector
from mbed_tools.devices._internal.candidate_device import CandidateDevice, FilesystemMountpointError


logger = logging.getLogger(__name__)


class LinuxDeviceDetector(DeviceDetector):
    """Linux specific implementation of device detection."""

    def find_candidates(self) -> List[CandidateDevice]:
        """Return a list of CandidateDevices."""
        context = pyudev.Context()
        candidates = []
        for disk in context.list_devices(subsystem="block", ID_BUS="usb"):
            serial_number = disk.properties.get("ID_SERIAL_SHORT")
            try:
                candidates.append(
                    CandidateDevice(
                        mount_points=_find_fs_mounts_for_device(disk.properties.get("DEVNAME")),
                        product_id=disk.properties.get("ID_MODEL_ID"),
                        vendor_id=disk.properties.get("ID_VENDOR_ID"),
                        serial_number=serial_number,
                        serial_port=_find_serial_port_for_device(serial_number),
                    )
                )
            except FilesystemMountpointError:
                logger.warning(
                    f"A USB block device was detected at path {disk.properties.get('DEVNAME')}. However, the"
                    " file system has failed to mount. Please disconnect and reconnect your device and try again."
                    "If this problem persists, try running fsck.vfat on your block device, as the file system may be "
                    "corrupted."
                )
                continue
        return candidates


def _find_serial_port_for_device(disk_serial_id: str) -> Optional[str]:
    """Try to find a serial port associated with the given device."""
    for tty_dev in pyudev.Context().list_devices(subsystem="tty"):
        if tty_dev.properties.get("ID_SERIAL_SHORT") == disk_serial_id:
            return cast(str, tty_dev.properties.get("DEVNAME"))
    return None


def _find_fs_mounts_for_device(device_file_path: str) -> Tuple[Path, ...]:
    """Find the file system mount point for a block device file path."""
    return tuple(Path(part.mountpoint) for part in psutil.disk_partitions() if part.device == device_file_path)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Detect Mbed devices connected to host computer."""
import platform
from typing import Iterable

from mbed_tools.devices._internal.candidate_device import CandidateDevice
from mbed_tools.devices._internal.base_detector import DeviceDetector
from mbed_tools.devices.exceptions import UnknownOSError


def detect_candidate_devices() -> Iterable[CandidateDevice]:
    """Returns Candidates connected to host computer."""
    detector = _get_detector_for_current_os()
    return detector.find_candidates()


def _get_detector_for_current_os() -> DeviceDetector:
    """Returns DeviceDetector for current operating system."""
    if platform.system() == "Windows":
        from mbed_tools.devices._internal.windows.device_detector import WindowsDeviceDetector

        return WindowsDeviceDetector()
    if platform.system() == "Linux":
        from mbed_tools.devices._internal.linux.device_detector import LinuxDeviceDetector

        return LinuxDeviceDetector()
    if platform.system() == "Darwin":
        from mbed_tools.devices._internal.darwin.device_detector import DarwinDeviceDetector

        return DarwinDeviceDetector()

    raise UnknownOSError(
        f"We have detected the OS you are running is '{platform.system()}'. "
        "Unfortunately we haven't implemented device detection support for this OS yet. Sorry!"
    )



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Flash binary onto the connected device."""

import shutil
import os
import pathlib
import platform

from mbed_tools.build.exceptions import BinaryFileNotFoundError


def _flash_dev(disk: pathlib.Path, image_path: pathlib.Path) -> None:
    """Flash device using copy method.

    Args:
        disk: Device mount point.
        image_path: Image file to be copied to device.
    """
    shutil.copy(image_path, disk, follow_symlinks=False)
    if not platform.system() == "Windows":
        os.sync()


def _build_binary_file_path(program_path: pathlib.Path, build_dir: pathlib.Path, hex_file: bool) -> pathlib.Path:
    """Build binary file name.

    Args:
       program_path: Path to the Mbed project.
       build_dir: Path to the CMake build folder.
       hex_file: Use hex file.

    Returns:
        Path to binary file.

    Raises:
        BinaryFileNotFoundError: binary file not found in the path specified
    """
    fw_fbase = build_dir / program_path.name
    fw_file = fw_fbase.with_suffix(".hex" if hex_file else ".bin")
    if not fw_file.exists():
        raise BinaryFileNotFoundError(f"Build program file (firmware) not found {fw_file}")
    return fw_file


def flash_binary(
    mount_point: pathlib.Path, program_path: pathlib.Path, build_dir: pathlib.Path, mbed_target: str, hex_file: bool
) -> pathlib.Path:
    """Flash binary onto a device.

    Look through the connected devices and flash the binary if the connected and built target matches.

    Args:
       mount_point: Mount point of the target device.
       program_path: Path to the Mbed project.
       build_dir: Path to the CMake build folder.
       mbed_target: The name of the Mbed target to build for.
       hex_file: Use hex file.
    """
    fw_file = _build_binary_file_path(program_path, build_dir, hex_file)
    _flash_dev(mount_point, fw_file)
    return fw_file



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pytest

from mbed_tools.lib.python_helpers import flatten_nested


@pytest.mark.parametrize(
    "input_list, expected_result",
    (
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]),
        ([1, [2], 3, 4, 5], [1, 2, 3, 4, 5]),
        ([[1, 2, 3], [4, 5]], [1, 2, 3, 4, 5]),
        ([(1, 2, 3), (4, 5)], [1, 2, 3, 4, 5]),
        ([[1, [[2, [3]]]], [4, 5]], [1, 2, 3, 4, 5]),
        ([["alan", [["bob", ["sally"]]]], ["jim", "jenny"]], ["alan", "bob", "sally", "jim", "jenny"]),
    ),
)
def test_flatten_nested_list(input_list, expected_result):
    assert flatten_nested(input_list) == expected_result



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

import pytest

from unittest import mock

from mbed_tools.project._internal.libraries import MbedLibReference, LibraryReferences
from mbed_tools.project.exceptions import VersionControlError
from tests.project.factories import make_mbed_lib_reference


@pytest.fixture
def mock_clone():
    with mock.patch("mbed_tools.project._internal.git_utils.clone") as clone:
        yield clone


@pytest.fixture
def mock_checkout():
    with mock.patch("mbed_tools.project._internal.git_utils.checkout") as checkout:
        yield checkout


@pytest.fixture
def mock_fetch():
    with mock.patch("mbed_tools.project._internal.git_utils.fetch") as fetch:
        yield fetch


@pytest.fixture
def mock_get_repo():
    with mock.patch("mbed_tools.project._internal.git_utils.get_repo") as get_repo:
        yield get_repo


@pytest.fixture
def mock_get_default_branch():
    with mock.patch("mbed_tools.project._internal.git_utils.get_default_branch") as get_default_branch:
        yield get_default_branch


@pytest.fixture
def mock_repo():
    with mock.patch("mbed_tools.project._internal.git_utils.git.Repo") as repo:
        yield repo


class TestLibraryReferences:
    def test_hydrates_top_level_library_references(self, mock_clone, tmp_path):
        fs_root = pathlib.Path(tmp_path, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git")
        mock_clone.side_effect = lambda url, dst_dir, *args: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        mock_clone.assert_called_once_with(lib.get_git_reference().repo_url, lib.source_code_path)
        assert lib.is_resolved()

    def test_hydrates_recursive_dependencies(self, mock_clone, tmp_path):
        fs_root = pathlib.Path(tmp_path, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git")
        # Create a lib reference without touching the fs at this point, we want to mock the effects of a recursive
        # reference lookup and we need to assert the reference was resolved.
        lib2 = MbedLibReference(
            reference_file=(lib.source_code_path / "lib2.lib"), source_code_path=(lib.source_code_path / "lib2")
        )
        # Here we mock the effects of a recursive reference lookup. We create a new lib reference as a side effect of
        # the first call to the mock. Then we create the src dir, thus resolving the lib, on the second call.
        mock_clone.side_effect = lambda url, dst_dir, *args: (
            make_mbed_lib_reference(pathlib.Path(dst_dir), name=lib2.reference_file.name, ref_url="https://valid2"),
            lib2.source_code_path.mkdir(),
        )

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        assert lib.is_resolved()
        assert lib2.is_resolved()

    def test_does_perform_checkout_of_default_repo_branch_if_no_git_ref_exists(
        self, mock_get_repo, mock_checkout, mock_fetch, mock_get_default_branch, mock_clone, tmp_path
    ):
        fs_root = pathlib.Path(tmp_path, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git", resolved=True)

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.checkout(force=False)

        mock_fetch.assert_called_once_with(mock_get_repo(), mock_get_default_branch())
        mock_checkout.assert_called_once_with(mock_get_repo(), "FETCH_HEAD", force=False)

    def test_performs_checkout_if_git_ref_exists(self, mock_get_repo, mock_checkout, mock_fetch, mock_clone, tmp_path):
        fs_root = pathlib.Path(tmp_path, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234", resolved=True)

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.checkout(force=False)

        mock_fetch.assert_called_once_with(mock_get_repo(), lib.get_git_reference().ref)
        mock_checkout.assert_called_once_with(mock_get_repo.return_value, "FETCH_HEAD", force=False)

    def test_fetch_does_not_perform_checkout_if_no_git_ref_exists(
        self, mock_get_repo, mock_checkout, mock_fetch, mock_clone, tmp_path
    ):
        fs_root = pathlib.Path(tmp_path, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git")
        mock_clone.side_effect = lambda url, dst_dir, *args: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        mock_fetch.assert_not_called()
        mock_checkout.assert_not_called()

    def test_fetch_performs_checkout_if_ref_is_hash(
        self, mock_get_repo, mock_clone, mock_fetch, mock_checkout, tmp_path
    ):
        num_times_called = 0

        def clone_side_effect(url, dst_dir, *args):
            nonlocal num_times_called
            if num_times_called == 0:
                num_times_called += 1
                raise VersionControlError("Failed to clone")
            elif num_times_called == 1:
                num_times_called += 1
                dst_dir.mkdir()
            else:
                assert False

        fs_root = pathlib.Path(tmp_path, "foo")
        lib = make_mbed_lib_reference(fs_root, ref_url="https://git#398bc1a63370")
        mock_clone.side_effect = clone_side_effect

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        mock_clone.assert_called_with(lib.get_git_reference().repo_url, lib.source_code_path)
        mock_fetch.assert_called_once_with(None, lib.get_git_reference().ref)
        mock_checkout.assert_called_once_with(None, "FETCH_HEAD")

    def test_raises_when_no_such_ref(self, mock_repo, mock_clone, mock_fetch, mock_checkout, tmp_path):
        num_times_called = 0

        def clone_side_effect(url, dst_dir, *args):
            nonlocal num_times_called
            if num_times_called == 0:
                num_times_called += 1
                raise VersionControlError("Failed to clone")
            elif num_times_called == 1:
                num_times_called += 1
                dst_dir.mkdir()
            else:
                assert False

        fs_root = pathlib.Path(tmp_path, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234")

        mock_clone.side_effect = clone_side_effect
        mock_fetch.side_effect = None
        mock_checkout.side_effect = VersionControlError("Failed to checkout")

        with pytest.raises(VersionControlError, match="Failed to checkout"):
            lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
            lib_refs.fetch()

    def test_doesnt_fetch_for_branch_or_tag(self, mock_clone, mock_fetch, mock_checkout, tmp_path):
        fs_root = pathlib.Path(tmp_path, "foo")
        make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234")

        mock_clone.side_effect = lambda url, dst_dir, *args: dst_dir.mkdir()

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        mock_fetch.assert_not_called()
        mock_checkout.assert_not_called()

    def test_does_not_resolve_references_in_ignore_paths(self, mock_get_repo, mock_checkout, mock_clone, tmp_path):
        fs_root = pathlib.Path(tmp_path, "mbed-os")
        make_mbed_lib_reference(fs_root, ref_url="https://git#lajdhalk234")

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.fetch()

        mock_clone.assert_not_called()

    def test_fetches_only_requested_ref(self, mock_repo, tmp_path):
        fs_root = pathlib.Path(tmp_path, "foo")
        fake_ref = "28eeee2b4c169739192600b92e7970dbbcabd8d0"
        make_mbed_lib_reference(fs_root, ref_url=f"https://git#{fake_ref}", resolved=True)

        lib_refs = LibraryReferences(fs_root, ignore_paths=["mbed-os"])
        lib_refs.checkout(force=False)

        mock_repo().git.fetch.assert_called_once_with("origin", fake_ref)
